# frozen_string_literal: true

require_relative 'search/version'
require 'set'

module Reflective
  module Search
    class Error < StandardError; end

    class Result
      attr_reader :id, :score

      def initialize(id:, score:)
        @id = id
        @score = score
      end
    end

    class Results
      attr_reader :matches

      def initialize(matches)
        @matches = matches
      end

      def ids
        matches.map(&:id)
      end
    end

    class Index
      def initialize(name, &block)
        @name = name
        @indexes = {}
        instance_eval(&block)
      end

      def category(name)
        @indexes[name] = Hash.new { Set.new }
      end

      def tokenize(value)
        value&.downcase&.split&.map(&:strip) || []
      end

      def add(item)
        @indexes.each do |category, index|
          tokenize(item.send(category)).each do |word|
            index[word] = index[word].add item.id
          end
        end
      end

      def clear
        @indexes.each do |name, _|
          @indexes[name] = Hash.new { Set.new }
        end
      end

      def search(query)
        tokenized_query = tokenize(query).to_set
        matches = @indexes.reduce([]) do |acc, (_, index)|
          index.select { |k, _| tokenized_query.include? k }
               .each_value do |found|
            acc += found.to_a
          end
          acc
        end
        to_results(matches)
      end

      private

      def to_results(matches)
        Results.new(matches.tally
                 .map { |id, count| Result.new(id: id, score: count) }
                 .sort { |a, b| b.score <=> a.score })
      end
    end
  end
end
